/* tableData.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "tableData.h"
#include "database.h"
#include "tableDataProxy.h"
#include "tableInfo.h"

#include <map>
#include <list>
#include <time.h>

// We keep around a list of cached TableData values for each table.
// This allows us to minimize hits on the database when we make
// several consecutive requests for the same data (as often happens
// when loading a CD record where each track links to the same
// artist).

// We only keep the cached value around for a brief time, so we won't
// miss values that were changed at another station elsewhere.  Also,
// we don't want the cache to grow absurdly large.

// This class manages that cache.

static const time_t max_keep_seconds = 30;

class TableDataCache {
public:
  TableData *find_cache_entry(TableID table_id, int primary_key_value);
  TableData *make_cache_entry(TableID table_id, int primary_key_value);
  TableData *update_cache_entry(const TableData &copy);

private:
  class Entry {
  public:
    Entry(TableID table_id, int primary_key_value);
    Entry(const TableData &copy);

    time_t _time;
    TableData _data;
  };

  typedef list<Entry> Entries;
  typedef map<TableID, Entries> Cache;
  Cache _cache;
};

static TableDataCache _table_cache;

TableDataCache::Entry::
Entry(TableID table_id, int primary_key_value) :
  _data(table_id),
  _time(time(NULL)) 
{
  _data.look_up(primary_key_value);
}

TableDataCache::Entry::
Entry(const TableData &copy) :
  _data(copy),
  _time(time(NULL)) 
{
}

TableData *TableDataCache::
find_cache_entry(TableID table_id, int primary_key_value) {
  Entries &entries = _cache[table_id];

  time_t now = time(NULL);

  // First, flush out the oldest entries in the cache.
  Entries::iterator ei;
  time_t last_time = 0;
  ei = entries.begin();
  while (ei != entries.end() && 
	 (now - (*ei)._time) > max_keep_seconds) {
    time_t this_time = (*ei)._time;
    //    assert(this_time >= last_time);
    last_time = this_time;

    ++ei;
  }

  entries.erase(entries.begin(), ei);
  
  // Now, look for our record.
  for (ei = entries.begin(); ei != entries.end(); ++ei) {
    if ((*ei)._data.get_primary_key() == primary_key_value) {
      return &((*ei)._data);
    }
    time_t this_time = (*ei)._time;
    //    assert(this_time >= last_time);
    last_time = this_time;
  }

  // Nope, nothing in the cache.
  return NULL;
}

TableData *TableDataCache::
make_cache_entry(TableID table_id, int primary_key_value) {
  // First, see if the record is already there.
  TableData *cache_entry = find_cache_entry(table_id, primary_key_value);

  if (cache_entry == NULL) {
    // It's not.  Make a new one.
    Entries &entries = _cache[table_id];
    entries.push_back(Entry(table_id, primary_key_value));
    cache_entry = &entries.back()._data;
  }

  return cache_entry;
}

TableData *TableDataCache::
update_cache_entry(const TableData &copy) {
  TableID table_id = copy.get_table_id();
  int primary_key_value = copy.get_primary_key();
  // First, see if the record is already there.
  TableData *cache_entry = find_cache_entry(table_id, primary_key_value);

  if (cache_entry == NULL) {
    // It's not.  Make a new one.
    Entries &entries = _cache[table_id];
    entries.push_back(Entry(copy));
    cache_entry = &entries.back()._data;
  } else {
    // It is.  Update it.
    (*cache_entry) = copy;
  }

  return cache_entry;
}


TableData::Value::
Value() {
}

TableData::Value::
Value(const DataCell &current, const DataCell &db) :
  _current(current), _db(db) {
}

bool TableData::Value::
has_changed() const {
  return (_current != _db);
}

void TableData::Value::
set_db_value(const DataCell &value) {
  _db = _current = value;
}



TableData::
TableData(TableID table_id) {
  _def = TableDef::get_table(table_id);
  _table_id = table_id;
  _primary_key_value = -1;
  _new_data = true;

  clear();
}

TableData::
TableData(const TableData &copy) :
  _values(copy._values),
  _table_id(copy._table_id),
  _primary_key_value(copy._primary_key_value),
  _new_data(copy._new_data),
  _def(copy._def) {
}

TableData &TableData::
operator = (const TableData &copy) {
  _values = copy._values;
  _table_id = copy._table_id;
  _primary_key_value = copy._primary_key_value;
  _new_data = copy._new_data;
  _def = copy._def;
  return *this;
}

TableData:: 
~TableData() {
  // We clear out the primary key on destruction, just for
  // cleanliness.
  set_primary_key(-1, true);
}

TableData *TableData::
find_table_data(TableID table_id, int primary_key_value) {
  TableData *cache = 
    _table_cache.make_cache_entry(table_id, primary_key_value);
  return new TableData(*cache);
}

TableID TableData::
get_table_id() const {
  return _table_id;
}

// It is possible to have an empty table definition, e.g. a table with
// no fields.  This is only likely when we are talking about an
// extension table.
bool TableData::
has_fields() const {
  return !_def->_fields.empty();
}

void TableData::
clear() {
  _new_data = true;
  _values.clear();
}

bool TableData::
look_up(int primary_key_value) {
  // This is a handy assert to have for now.  Although nothing in the
  // code assumes it anywhere, it should be impossible to get a
  // primary key of zero, since the Postgres sequence mechanism starts
  // counting at 1.  Thus, if we try to look up a row with a primary
  // key of zero, it's probably an error.
  assert(primary_key_value != 0);

  set_primary_key(-1, true);
  if (primary_key_value < 0) {
    clear();
    return false;
  }

  if (has_fields()) {
    // Now we need to build a select statement to read the table
    // data for this page.

    string field_names;
    TableDef::Fields::const_iterator fi;
    for (fi = _def->_fields.begin(); fi != _def->_fields.end(); ++fi) {
      const TableField &field = (*fi).second;
      field_names += ", " + field._field_name;
    }

    string table_key = get_table_key(_table_id);

    string select = 
      "SELECT " + field_names.substr(2) + " FROM " + 
      get_table_name(_table_id);

    if (!table_key.empty()) {
      select += " WHERE " +
	table_key + "=" + DataCell(primary_key_value);
    }
    
    Database *db = get_table_db(_table_id);
    QueryResult qr = db->sql_command(select);
    
    if (qr.get_num_rows() == 0) {
      // No table data for this primary key.
      clear();

    } else {
      // Each column in the result corresponds to a field.
      TableDef::Fields::const_iterator fi;
      int c = 0;
      for (fi = _def->_fields.begin(); fi != _def->_fields.end(); ++fi) {
	int field_id = (*fi).first;
	const TableField &field = (*fi).second;

	set_db_value(field_id, qr.get_cell(0, c));
	c++;
      }
      
      set_primary_key(primary_key_value, false);
    }
  }

  return !_new_data;
}

void TableData::
set_primary_key(int primary_key_value, bool new_data) {
  _primary_key_value = primary_key_value;
  _new_data = new_data;
}

int TableData::
get_primary_key() const {
  return _primary_key_value;
}

bool TableData::
is_new_data() const {
  return _new_data;
}

int TableData::
record_db() {
  if (!has_fields()) {
    return _primary_key_value;
  }

  string table_key = get_table_key(_table_id);
  Database *db = get_table_db(_table_id);

  // If the table has no primary key (e.g. the configure table), we
  // always update the data with an UPDATE command, never INSERT.  The
  // data is never "new".
  if (table_key.empty()) {
    _new_data = false;
    _primary_key_value = 1;
  }

  if (_new_data) {
    if (_primary_key_value < 0) {
      // Get a new key value.
      int new_key = 
	db->get_next_sequence(get_table_key_sequence(_table_id));
      set_primary_key(new_key, true);
    }

    // Build up an insert command.
    string field_names;
    string field_values;

    TableDef::Fields::const_iterator fi;
    for (fi = _def->_fields.begin(); fi != _def->_fields.end(); ++fi) {
      int field_id = (*fi).first;
      const TableField &field = (*fi).second;

      // Since we'll be handling the primary key explicitly, we don't
      // need to add it to the strings.
      if (field._field_name != table_key) {
	DataCell &cell = (*this)[field_id];

	// If the field references a foreign table, also update the
	// foreign table.  This incidentally guarantees that we have a
	// usable foreign key (i.e. an integer, not -1).
	cell.record_db();

	field_names += ", " + field._field_name;
	field_values += ", " + cell;
      }
    }

    string insert =
      "INSERT INTO " + get_table_name(_table_id) +
      "(" + get_table_key(_table_id) + field_names +
      ") VALUES (" + DataCell(_primary_key_value) + field_values + ")";

    db->sql_command(insert);

    _new_data = false;
  } else {
    // Build up an update command.
    string field_assign;

    Values::iterator vi;
    for (vi = _values.begin(); vi != _values.end(); ++vi) {
      int field_id = (*vi).first;
      Value &value = (*vi).second;
      DataCell &cell = value._current;

      // If the field references a foreign table, also update the
      // foreign table.  This incidentally guarantees that we have a
      // usable foreign key (i.e. an integer, not -1).
      cell.record_db();

      if (value.has_changed()) {
	const TableField &field = _def->get_field(field_id);
	field_assign += ", " + field._field_name + "=" + cell;
      }
    }

    if (!field_assign.empty()) {
      string update =
	"UPDATE " + get_table_name(_table_id) +
	" SET " + field_assign.substr(2);

      if (!table_key.empty()) {
	update += 
	  " WHERE " +
	  table_key + "=" + DataCell(_primary_key_value);
      }

      db->sql_command(update);
    }
  }

  // Now update our internal bookkeeping to show that the database now
  // reflects our current data, and we'll be able to tell which values
  // have changed next time.
  Values::iterator vi;
  for (vi = _values.begin(); vi != _values.end(); ++vi) {
    (*vi).second._db = (*vi).second._current;
  }

  // Finally, update the cache.
  _table_cache.update_cache_entry(*this);

  return _primary_key_value;
}

int TableData::
record_db(int primary_key_value) {
  if (primary_key_value != get_primary_key()) {
    set_primary_key(primary_key_value, true);
  }
  return record_db();
}


DataCell &TableData::
operator [] (int field_id) {
  Values::iterator vi;
  vi = _values.find(field_id);
  if (vi != _values.end()) {
    return (*vi).second._current;

  } else {
    // The field_id value does not exist; create it with the default
    // value.
    Value &value = _values[field_id];

    TableDef::Fields::const_iterator fi = _def->_fields.find(field_id);
    if (fi != _def->_fields.end()) {
      const TableField &field = (*fi).second;
      value._current = field.get_default_value();
    }

    return value._current;
  }

}

const DataCell TableData::
operator [] (int field_id) const {
  // We have to be careful not to add any new rows if the field_id
  // does not exist.
  Values::const_iterator vi;
  vi = _values.find(field_id);
  if (vi != _values.end()) {
    return (*vi).second._current;
  } else {
    // The field_id value does not exist; look up the default value.

    TableDef::Fields::const_iterator fi = _def->_fields.find(field_id);
    if (fi != _def->_fields.end()) {
      const TableField &field = (*fi).second;
      return field.get_default_value();
    } else {
      return DataCell();
    }
  }
}

void TableData::
set_db_value(int field_id, const DataCell &value) {
  _values[field_id].set_db_value(value);
}

const DataCell TableData::
get_db_value(int field_id) const {
  // We have to be careful not to add any new rows if the field_id
  // does not exist.
  Values::const_iterator vi;
  vi = _values.find(field_id);
  if (vi != _values.end()) {
    return (*vi).second._db;
  } else {
    // The field_id value does not exist; look up the default value.

    TableDef::Fields::const_iterator fi = _def->_fields.find(field_id);
    if (fi != _def->_fields.end()) {
      const TableField &field = (*fi).second;
      return field.get_default_value();
    } else {
      return DataCell();
    }
  }
}

void TableData::
output(ostream &out) const {
  if (is_new_data()) {
    out << "table " << _table_id << ", new record";
  } else {
    out << "table " << _table_id << ", key " << get_primary_key();
  }
}

    
